/* -*- Mode: C -*-
 *Header:
 *File: object.C (for C++ & cc)
 *Author: Noda Itsuki
 *Date: 1995/02/21
 *EndHeader:
 */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This file is a part of SoccerServer.

    This code is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */


#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
using namespace std;

#include "param.h"
#include "types.h"
#include "utility.h"
#include "object.h"
#include "field.h"


/*
 *===================================================================
 *Part: Plane Vector
 *===================================================================
 */
PVector Polar2PVector(Value r, Angle ang)
{
	return PVector(r * cos(ang), r * sin(ang)) ;
}

PVector::PVector(const double vx, const double vy)
{
	x = vx ;
	y = vy ;
}

void PVector::operator +=(const PVector& v)
{
	x += v.x ;
	y += v.y ;
}

void PVector::operator -=(const PVector& v)
{
	x -= v.x ;
	y -= v.y ;
}

void PVector::operator *=(const double& a)
{
	x *= a ;
	y *= a ;
}

void PVector::operator /=(const double& a)
{
	x /= a ;
	y /= a ;
}

ostream& operator<< (ostream& o, const PVector& v) 
{
	return o << "#V[" << v.x << "," << v.y << "]" ;
}

PVector operator +(const PVector& a, const PVector& b)
{
	return PVector((a.x + b.x), (a.y + b.y)) ;
}

PVector operator -(const PVector& a, const PVector& b)
{
	return PVector((a.x - b.x), (a.y - b.y)) ;
}

void PVector::normalize(const Value& l) 
{
	*this *= (l/max(r(),EPS)) ;
}

Value PVector::distance(const PVector& orig)
{
	return (*this - orig).r() ;
}

Angle PVector::angle()
{
	return Atan(y,x) ;
}

Angle PVector::angle(const PVector& dir)
{
	Angle ang = Atan(dir.y,dir.x) - Atan(y,x) ;

	return normalize_angle(ang) ;
}

void PVector::rotate(const Angle& ang)
{
	Value r1 = r() ;
	Angle th1 = th() ;

	x = r1 * cos(th1 + ang) ;
	y = r1 * sin(th1 + ang) ;
}

Angle PVector::vangle(const PVector& target, const PVector& origin) 
{
	PVector axis = origin - *this ;
	PVector dir = target - *this ;
	return axis.angle(dir) ;
}

Angle PVector::vangle(const PVector& target, const Angle& origin)
{
	PVector dir = target - *this ;
	Angle ang = dir.angle() - origin ;

	return normalize_angle(ang) ;
}

/*
 *===================================================================
 *Part: Area
 *===================================================================
 */

RArea::RArea(const Value l, const Value r, const Value t, const Value b) 
{
	left = l ; right = r ;
	top = t ; bottom = b ;
}

RArea::RArea(const PVector center,const PVector size)
{
	left = center.x - size.x/2.0 ;
	right = center.x + size.x/2.0 ;
	top = center.y - size.y/2.0 ;
	bottom = center.y + size.y/2.0 ;
}

Logical RArea::inArea(const PVector &p)
{
	return (p.x >= left) && (p.x <= right) && (p.y >= top) && (p.y <= bottom) ;
}

PVector RArea::nearestHEdge(const PVector &p) 
		/* find nearest horizontal line */
{
	static PVector r ;

	r.x = min(max(p.x,left),right) ;
	r.y = (Abs(p.y - top) < Abs(p.y - bottom)) ? top : bottom ;

	return r ;
}

PVector RArea::nearestVEdge(const PVector &p) 
		/* find nearest vertical line */
{
	static PVector r ;

	r.x = (Abs(p.x - left) < Abs(p.x - right)) ? left : right ;
	r.y = min(max(p.y,top),bottom) ;

	return r ;
}

PVector RArea::nearestEdge(const PVector &p) 
{
	if(min(Abs(p.x-left),Abs(p.x-right)) < min(Abs(p.y-top),Abs(p.y-bottom)))
		return nearestVEdge(p) ;
	else
		return nearestHEdge(p) ;
}

PVector RArea::randomize() 
{
	static PVector r ;

	r.x = drand(left, right) ;
	r.y = drand(bottom, top) ;

	return r ;
}

ostream& operator<< (ostream& o, const RArea& a) 
{
	return o << "#A[h:" << a.left << "~" << a.right 
			<< ",v:" << a.top << "~" << a.bottom << "]" ;
}

CArea::CArea(const PVector cnt, const Value radius)
{
	center = cnt ;
	r = radius ;
}

Logical CArea::inArea(const PVector &p)
{
	return r >= center.distance(p) ;
}

PVector CArea::nearestEdge(const PVector &p)
{
	PVector dif = p - center ;
	if (dif.x == 0.0 && dif.y == 0.0)
		dif = PVector(EPS, EPS) ;

	dif.normalize(r) ;

	return center + dif ;
}

ostream& operator<< (ostream& o, const CArea& a) 
{
	return o << "#A[x:" << a.center.x <<
				",y:" << a.center.y <<
				",r:" << a.r << "]" ;
}

/*
 *===================================================================
 *Part: Visual Info
 *===================================================================
 */

VisualInfo::VisualInfo(double version_tmp)
{
	object = (PObject *)NULL ;
	version = version_tmp ;
}

VisualInfo::VisualInfo()
{
        object = (PObject *)NULL ;
}

void VisualInfo::SetVersion(double version_tmp)
{
        version = version_tmp;
}

ostream& operator<< (ostream& o, const VisualInfo& vi)
{
	Value ed ;
	Value rdist_change ;
	Angle rdir_change ;
	char tmp[128] ;
				/* modified distance */
	ed = Quantize(exp(Quantize(log(vi.distance + EPS),vi.qstep)),0.1) ;

	sscanf(vi.name,"(%s",tmp) ;
	if (!strcmp(tmp,"player")) {
		double prob ;
		prob = (ed-TEAM_FAR_LENGTH)/(TEAM_TOOFAR_LENGTH-TEAM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
		  /* th 19.05.00 */
		  if (vi.version >= 6.0)
		    o << "(" << ((Player *)(vi.object))->short_name_toofar << " " ;
		  else
			o << "(" << ((Player *)(vi.object))->name_toofar << " " ;
			if (vi.highquality)
				o << ed << " " ; 
			return o << ((vi.version >= 7.0) ?
				     Rad2IDegRound(vi.direction) :
				     Rad2IDeg(vi.direction))
				 << ")" ;
		}

		prob = (ed-UNUM_FAR_LENGTH)/(UNUM_TOOFAR_LENGTH-UNUM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
		  /* th 19.05.00 */
		  if (vi.version >= 6.0)
		    o << "(" << ((Player *)(vi.object))->short_name_far << " " ;
		  else
			o << "(" << ((Player *)(vi.object))->name_far << " " ;
			if (vi.highquality)
				o << ed << " " ;
			return o << ((vi.version >= 7.0) ?
				     Rad2IDegRound(vi.direction) :
				     Rad2IDeg(vi.direction))
				 << ")" ;
		}
		/* th 19.05.00 */
		if (vi.version >= 6.0)
		  o << "(" << vi.short_name << " " ;
		else
		o << "(" << vi.name << " " ;

		if (vi.highquality) {
			o << ed << " " ;
			if (vi.distance != Zero)
				rdist_change = ed * Quantize(vi.dist_change/vi.distance, 0.02) ;
			else
				rdist_change = Zero ;
			if (vi.dir_change != Zero)
#ifndef NEW_QSTEP
				rdir_change = Quantize(vi.dir_change, 0.1) ;
#else
				rdir_change = Quantize(vi.dir_change,vi.qstep_dir) ;
#endif
			else
				rdir_change = Zero ;
			if (vi.version >= 7.0)
				return o << Rad2IDegRound(vi.direction) << " "
					 << rdist_change << " "
					 << rdir_change << " "
					 << Rad2IDegRound(vi.body_dir) << " "
					 << Rad2IDegRound(vi.face_dir) << ")" ;
			if (vi.version >= 5.0)
				return o << Rad2IDeg(vi.direction) << " "
					 << rdist_change << " "
					 << rdir_change << " "
					 << Rad2IDeg(vi.body_dir) << " "
					 << Rad2IDeg(vi.face_dir) << ")" ;
			if (vi.version > 4.0)
				return o << Rad2IDeg(vi.direction)  << " "
					 << rdist_change << " " 
					 << rdir_change << " "
					 << Rad2IDeg(vi.body_dir) << ")" ;
			if (vi.version > 3.0)
				return o << Rad2IDeg(vi.direction) << " "
					 << rdist_change << " " 
					 << rdir_change << " "
					 << Rad2IDeg(vi.body_dir) << ")" ;
			else
				return o << Rad2IDeg(vi.direction) << " "
					 << rdist_change << " " 
					 << rdir_change << ")" ;
		}
		else
			return o << ((vi.version >= 7.0) ?
				     Rad2IDegRound(vi.direction) :
				     Rad2IDeg(vi.direction))
				 << ")" ;
	}
	else if (!strcmp(tmp, "line")) {
	  /* th 19.05.00 */
	  if (vi.version >= 6.0)
	    o << "(" << vi.short_name << " " ;
	  else
		o << "(" << vi.name << " " ;
		if (vi.highquality)
			o << ed << " " ;
		return o << ((vi.version >= 7.0) ?
			     Rad2IDegRound(vi.direction) :
			     Rad2IDeg(vi.direction))
			 << ")" ;
	}
	else if (!strcmp(tmp, "ball)") || !strcmp(tmp, "flag")) {
		double prob ;
		prob = (ed-UNUM_FAR_LENGTH)/(UNUM_TOOFAR_LENGTH-UNUM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
		  /* th 19.05.00 */
		  if (vi.version >= 6.0)
		    o << "(" << vi.short_name << " " ;
		  else
			o << "(" << vi.name << " " ;
		  if (vi.highquality)
		    o << ed << " " ;
		  return o << ((vi.version >= 7.0) ?
			       Rad2IDegRound(vi.direction) :
			       Rad2IDeg(vi.direction))
			   << ")" ;
		}
	}
	else {		/* near object that can't see */
	  /* th 19.05.00 */
	  if (vi.version >= 6.0)
	    o << "(" << vi.short_name << " " ;
	  else
		o << "(" << vi.name << " " ;
	  if (vi.highquality)
	    o << ed << " " ;
	  return o << ((vi.version >= 7.0) ?
		       Rad2IDegRound(vi.direction) :
		       Rad2IDeg(vi.direction))
		   << ")" ;
	}
	/* th 19.05.00 */
	if (vi.version >= 6.0)
	  o << "(" << vi.short_name << " " ;
	else
	o << "(" << vi.name << " " ;
	if (vi.highquality) {
		o << ed << " " ;
		if (vi.distance != Zero)
			rdist_change = ed * Quantize(vi.dist_change/vi.distance, 0.02) ;
		else
			rdist_change = Zero ;
		if (vi.dir_change != Zero)
#ifndef	NEW_QSTEP
			rdir_change = Quantize(vi.dir_change, 0.1) ;
#else
			rdir_change = Quantize(vi.dir_change, vi.qstep_dir) ;
#endif
		else
			rdir_change = Zero ;
		return o << ((vi.version >= 7.0) ?
			     Rad2IDegRound(vi.direction) :
			     Rad2IDeg(vi.direction))
			 << " "
			 << rdist_change << " " << rdir_change << ")" ;
	}
	else
		return o << ((vi.version >= 7.0) ?
			     Rad2IDegRound(vi.direction) :
			     Rad2IDeg(vi.direction))
			 << ")" ;
}

/*
 *===================================================================
 *Part: PObject 
 *===================================================================
 */

TheNumber PObject::Max = 0 ;

/* pfr 06/07/200 added short name support */
PObject::PObject(const Name nm, const Name snm,
		 const PVector& p, const Value v) 
{
	id = Max ;
	Max++ ;
	name = nm ;
	/* th 19.05.00 */
	short_name = snm ;
	stadium = (Stadium *)NULL ;
	vi = (VisualInfo *)NULL ;
	pos = p ;
	obj_ver = v ;
	size = 0.0 ;
	angle = 0.0 ;
	enable = TRUE ;
}

ostream& operator<< (ostream& o, const PObject& obj) 
{
	o << "#Ob[" << obj.id ;
	if( obj.name != NULLCHAR )
		o << ":" << obj.name << "" ;
	return o 
		<< ":pos=" << obj.pos 
		<< ",size=" << obj.size 
		<< ",angle=" << obj.angle
		<< "]" ;
}

void PObject::Set(const PVector& p, const Value& s)
{
	pos = p ;
	size = s ;
}

Angle PObject::vangle(const PObject& obj)
{
	return pos.vangle(obj.pos,angle) ;
}

int Player::scanField()
{
	PObjectTable *tbl ;
	MPObjectTable *mtbl ;
	int i,j,c ;
	Angle ang, angle_sight ;
	char tmp[16] ;
	PVector etmp, vtmp ;

	if(!vi){ 
		//vi = new VisualInfo[MaxObject](version) ;
		vi = new VisualInfo[MaxObject];
		for (i = 0; i < MaxObject; i++)
		{
			vi[i].SetVersion(version);
		}

#ifdef NEW_QSTEP
		for ( i = 0; i < MaxObject; i++)
			vi[i].qstep_dir = this->dir_qstep_player ;
#endif
	}
	c = 0 ;

	tbl = &(stadium->votable) ;
	mtbl = &(stadium->motable) ;

	angle_sight = angle + angle_neck_committed ;

	for(i = 0 ; i < tbl->n ; i++ ) {
		if(tbl->object[i] == this || (*tbl->object[i]).enable == FALSE)
			continue ;
		ang = this->vangle(*(tbl->object[i])) - angle_neck_committed ;
		ang = normalize_angle(ang) ;
		if(fabs(ang) < vis_angle/2.0) {
			vi[c].highquality = highquality ;
			vi[c].object = tbl->object[i] ;
			vi[c].name = tbl->object[i]->name ;
			/* th 19.05.00 */
			if (version >= 6.0)
			  vi[c].short_name = tbl->object[i]->short_name ;
			vi[c].distance = this->pos.distance(tbl->object[i]->pos) ;
			vi[c].direction = ang ;

			if (vi[c].distance == Zero)
				vi[c].dist_change = vi[c].dir_change = Zero ;
			else {
				etmp = tbl->object[i]->pos - this->pos ;
				sscanf(tbl->object[i]->name, "(%[^ )]", tmp);
				if (!strcmp(tmp, "ball")) {
#ifndef	NEW_QSTEP
					vi[c].qstep = stadium->server_param->dist_qstep ;
#else
					vi[c].qstep = this->dist_qstep_player ;
#endif					
					for(j = 0 ; j < mtbl->n ; j++ ) {
						if (tbl->object[i]->id == mtbl->object[j]->id) {
							vtmp = mtbl->object[j]->vel - this->vel ;
							break ;
						}
					}
				}
				else if (!strcmp(tmp, "player")) {
#ifndef	NEW_QSTEP
					vi[c].qstep = stadium->server_param->dist_qstep ;
#else
					vi[c].qstep = this->dist_qstep_player ;
#endif					
					for(j = 0 ; j < mtbl->n ; j++ ) {
						if (tbl->object[i]->id == mtbl->object[j]->id) {
							vtmp = mtbl->object[j]->vel - this->vel ;
							break ;
						}
					}
					vi[c].body_dir = 
						normalize_angle(tbl->object[i]->angle - (this->angle + this->angle_neck_committed)) ;
					vi[c].face_dir = 
						normalize_angle(
							(((Player *)tbl->object[i])->angle_neck_committed
								+ ((Player *)tbl->object[i])->angle)
								- (this->angle + this->angle_neck_committed)) ;

				}
				else {
#ifndef	NEW_QSTEP
					vi[c].qstep = stadium->server_param->land_qstep ;
#else
					vi[c].qstep = this->land_qstep_player ;
#endif
					vtmp.x = -this->vel.x ;
					vtmp.y = -this->vel.y ;
				}
				etmp /= vi[c].distance ;
				vi[c].dist_change = vtmp.x * etmp.x + vtmp.y * etmp.y ;
				vi[c].dir_change = RAD2DEG * (vtmp.y * etmp.x
								- vtmp.x * etmp.y) / vi[c].distance ;
			}
			c++ ;
		}
		else if (this->pos.distance(tbl->object[i]->pos) <= vis_distance){
			vi[c].highquality = highquality ;
			vi[c].object = tbl->object[i] ;
			sscanf(tbl->object[i]->name, "(%s", tmp);
			if (!strcmp(tmp, "flag")) {
				vi[c].name = O_TYPE_FLAG_NAME ;
				/* th 19.05.00 */
				if (version >= 6.0)
				  vi[c].short_name = O_TYPE_FLAG_NAME_SHORT ;
#ifndef NEW_QSTEP
				vi[c].qstep = stadium->server_param->land_qstep ;
#else
				vi[c].qstep = this->land_qstep_player ;
#endif
			}
			else if (!strcmp(tmp, "goal")) {
				vi[c].name = O_TYPE_GOAL_NAME ;
				/* th 19.05.00 */
				if (version >= 6.0)
				  vi[c].short_name = O_TYPE_GOAL_NAME_SHORT ;
#ifndef NEW_QSTEP
				vi[c].qstep = stadium->server_param->land_qstep ;
#else
				vi[c].qstep = this->land_qstep_player ;
#endif
			}
			else if (!strcmp(tmp, "ball)")) {
				vi[c].name = O_TYPE_BALL_NAME ;
				/* th 19.05.00 */
				if (version >= 6.0)
				  vi[c].short_name = O_TYPE_BALL_NAME_SHORT ;
#ifndef	NEW_QSTEP
				vi[c].qstep = stadium->server_param->dist_qstep ;
#else
				vi[c].qstep = this->dist_qstep_player ;
#endif					
			}
			else if (!strcmp(tmp, "player")) {
				vi[c].name = O_TYPE_PLAYER_NAME ;
				/* th 19.05.00 */
				if (version >= 6.0)
				  vi[c].short_name = O_TYPE_PLAYER_NAME_SHORT ;
#ifndef	NEW_QSTEP
				vi[c].qstep = stadium->server_param->dist_qstep ;
#else
				vi[c].qstep = this->dist_qstep_player ;
#endif					
			}
			else
			  {
				vi[c].name = "unknown" ;
				/* th 19.05.00 */
				if (version >= 6.0)
				  vi[c].short_name = "unknown" ;
			  }
			
			vi[c].distance = this->pos.distance(tbl->object[i]->pos) ;
			vi[c].direction = ang ;
			c++ ;
		}
	}

	// line information
	ang = (version >= 7.0 ) ?
	  Rad2IDegRound(normalize_angle(angle_sight)) :
	  Rad2IDeg(normalize_angle(angle_sight)) ;
	PObject *dir[2], *dir2[2] ;
	Value edge[2], r[4] ;
	int out_of_field = TRUE ;
	int even = FALSE ;

#ifndef WRONG_LINE_VIEW		/* line view type is normal */
	if (pos.y < -PITCH_WIDTH/2) {	/* above line_t */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
			}
			else if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
			}
			else if (ang == 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_t) ;
				if (ang < 90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_t) ;
				dir2[i] = &(stadium->field.line_b) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;
		}
	}
	else if (pos.y > PITCH_WIDTH/2) {	/* under line_b */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > -180 && ang < -90) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
			}
			else if (ang > -180 && ang < -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
			}
			else if (ang == -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_b) ;
				if (ang > -90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_b) ;
				dir2[i] = &(stadium->field.line_t) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;

		}
	}
	else if (pos.x < -PITCH_LENGTH/2) {	/* left of line_l */
		if (ang > -90 && ang < 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_t
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else if (ang > 0 && ang < 90) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;   // to line_b
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else if (ang == 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_l) ;
			if (ang < 0)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_l) ;
			dir2[i] = &(stadium->field.line_r) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else if (pos.x > PITCH_LENGTH/2) {	/* right of line_r */
		if (ang > -180 && ang < -90) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_t
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else if (ang > 90 && ang < 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;   // to line_b
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else if (ang == 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_r) ;
			if (ang < -90)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_r) ;
			dir2[i] = &(stadium->field.line_l) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else {		/* In the field */
		out_of_field = FALSE ;
		if (ang > -90 && ang <= 90) {
			dir[0] = &(stadium->field.line_r) ;
			edge[0] = PITCH_LENGTH/2 ;
		}
		else {
			dir[0] = &(stadium->field.line_l) ;
			edge[0] = -PITCH_LENGTH/2 ;
		}
		if (ang > -180 && ang <= 0) {
			dir[1] = &(stadium->field.line_t) ;
			edge[1] = -PITCH_WIDTH/2 ;
		}
		else {
			dir[1] = &(stadium->field.line_b) ;
			edge[1] = PITCH_WIDTH/2 ;
		}

		r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle_sight)) ;
		r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle_sight)) ;

		i = (r[0] <= r[1]) ? 0 : 1 ;
	}
#else		/* defined WRONG_LINE_VIEW */
	out_of_field = FALSE ;
	if (ang > -90 && ang <= 90) {
		dir[0] = &(stadium->field.line_r) ;
		edge[0] = PITCH_LENGTH/2 ;
	}
	else {
		dir[0] = &(stadium->field.line_l) ;
		edge[0] = -PITCH_LENGTH/2 ;
	}
	if (ang > -180 && ang <= 0) {
		dir[1] = &(stadium->field.line_t) ;
		edge[1] = -PITCH_WIDTH/2 ;
	}
	else {
		dir[1] = &(stadium->field.line_b) ;
		edge[1] = PITCH_WIDTH/2 ;
	}

	r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle_sight)) ;
	r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle_sight)) ;
	i = (r[0] <= r[1]) ? 0 : 1 ;
#endif

	if (i == 0)
		ang = -angle_sight + ((angle_sight > 0) ? PI/2 : -PI/2) ;
	else
		ang = -angle_sight ;
	if (ang > PI/2) ang -= PI ;
	if (ang < -PI/2) ang += PI ;

	vi[c].highquality = highquality ;
	vi[c].object = dir[i] ;
	vi[c].name = dir[i]->name ;
	/* th 19.05.00 */
	if (version >= 6.0)
	  vi[c].short_name = dir[i]->short_name ;
	vi[c].distance = r[i] ;
	vi[c].direction = ang ;
#ifndef NEW_QSTEP
	vi[c].qstep = stadium->server_param->land_qstep ;
#else
	vi[c].qstep = this->land_qstep_player ;
#endif
	c++ ;

	if (out_of_field) {
		vi[c].highquality = highquality ;
		vi[c].object = dir2[i] ;
		vi[c].name = dir2[i]->name ;
		/* th 19.05.00 */
		if (version >= 6.0)
		  vi[c].short_name = dir2[i]->short_name ;
		vi[c].distance = r[3-i] ;
#ifndef NEW_QSTEP
		vi[c].qstep = stadium->server_param->land_qstep ;
#else
		vi[c].qstep = this->land_qstep_player ;
#endif
		if (even)
			vi[c].direction = ang ;
		else
			vi[c].direction = ang + ((ang > 0) ? -PI/2 : PI/2) ;
		c++ ;
	}

	return c ;

}


/*
 *===================================================================
 *Part: MPObject
 *===================================================================
 */

/* pfr 06/07/200 added short name support */
MPObject::MPObject(const Name nm, const Name s_nm) : PObject(nm, s_nm) { }

ostream& operator<< (ostream& o, const MPObject& obj) 
{
	o << "#Ob[" << obj.id ;
	if( obj.name != NULLCHAR )
		o << ":" << obj.name << "" ;
	return o 
		<< ":pos=" << obj.pos 
	 	<< ",size=" << obj.size
	 	<< ",angle=" << Rad2IDegRound(obj.angle)
		<< ",vel=" << obj.vel 
		<< ",acc=" << obj.acc 
		<< ",decay=" << obj.decay
		<< ",randp=" << obj.randp
		<< "]" ;
}

void MPObject::Set(
				const MPObjectType& t,
				const PVector& p,
				const Value& s,
				const Angle& ang, 
				const PVector& v,
				const PVector& a,
				const Value& d,
				const Value& r)
{
	type = t ;
	pos = p ;
	size = s ;
	angle = ang ;
	vel = v ;
	acc = a ;
	decay = d ;
	randp = r ;
}

PVector MPObject::noise() 
{
	Value maxrnd = randp * vel.r() ;
	return PVector(drand(-maxrnd,maxrnd),drand(-maxrnd,maxrnd)) ;
}

PVector MPObject::wind()
{
	return PVector(vel.r() * (weather->wind_vector.x +
			drand(-weather->wind_rand, weather->wind_rand) ) /
			(weight * WIND_WEIGHT), vel.r() * (weather->wind_vector.y +
			drand(-weather->wind_rand, weather->wind_rand)) /
			(weight * WIND_WEIGHT)) ;
}

void MPObject::_inc() 
{
	Value tmp ;

	if (acc.x || acc.y)
	  {
	    double max_a = max_accel;
	    double max_s = max_speed;
	    if (this->type == MPO_Player)
	      {
		max_a /= stadium->slowdownOnTop ( pos, ((Player *)this)->team->side);
	        max_s /= stadium->slowdownOnTop ( pos, ((Player *)this)->team->side);
	      }

	    if ((tmp = acc.r()) > max_a )
	      acc *= (max_a / tmp);
	    vel += acc ;
	    if ((tmp = vel.r()) > max_s)
	      vel *= (max_s / tmp) ;
	}
	if (this->type == MPO_Player) {
                angle = ((Player *)this)->angle_body ;
                ((Player *)this)->angle_neck_committed = ((Player *)this)->angle_neck ;
	}
	vel += noise() ;
	vel += wind() ;
	pos += vel ;
	vel *= decay ;
	acc *= Zero ;
}

void MPObject::_turn()
{
    if (this->type == MPO_Player) {
                angle = ((Player *)this)->angle_body ;
                ((Player *)this)->angle_neck_committed = ((Player *)this)->angle_neck ;
	}
}

Logical MPObject::collisionP(const MPObject& obj) 
{
	if(&obj == this) return FALSE ;
	if(&obj.enable == FALSE) return FALSE  ;
	return this->pos.distance(obj.pos) < (size + obj.size) ;
}

void MPObject::collisionCheck(MPObjectTable& tbl)
{
	tbl.collisionCheck(this) ;
}

void MPObjectTable::collisionCheck(MPObject* obj)
{
	TheNumber i ;
	Logical colp = FALSE ;
	MPObject *ob ;

	for(i = 0 ; i < n ; i++) {
		ob = object[i] ;
		if(obj->collisionP(*ob)) {
			colp = TRUE ;
			obj->collide(*ob) ;
			if (obj->type == MPO_Ball && ob->type == MPO_Player) {
				Player *player = (Player *)ob ;
				player->alive |= BALL_TO_PLAYER ;
				player->alive |= BALL_COLLIDE ;
				/* we don't want to set last touch if the goalie has caught
				   the ball! */
				if (player->stadium->caught_ball == NEUTRAL)
				  player->stadium->last_touch = player ;
			}
			else if (obj->type == MPO_Player && ob->type == MPO_Ball) {
				Player *player = (Player *)obj ;
				player->alive |= PLAYER_TO_BALL ;
				player->alive |= BALL_COLLIDE ;				
				/* we don't want to set last touch if the goalie has caught
				   the ball! */
				if (player->stadium->caught_ball == NEUTRAL)
				  player->stadium->last_touch = player ;
			}
			else if ( obj->type == MPO_Player && ob->type == MPO_Player)
			  {
			    Player *player = (Player *)ob ;
			    player->alive |= PLAYER_COLLIDE ;
			    player = (Player *)obj ;
			    player->alive |= PLAYER_COLLIDE ;
			  }
		}
	}
	if(colp)
		obj->vel *= -0.1 ;
}

void MPObject::collide(MPObject& obj)
{
	Value r = size + obj.size ;
	PVector dif = (pos - obj.pos) ;
	Value d = pos.distance(obj.pos) ;
	Angle th = fabs(dif.angle(vel)) ;
	Value l1 = d * cos(th) ;
	Value h = d * sin(th) ;
	Value cosp = h / r ;
	Value sinp = sqrt(1.0 - square(cosp)) ;
	Value l2 = r * sinp ;
	PVector dv = vel ;

	dv.normalize(-(l1 + l2)) ;

	pos += dv ;
}

void MPObject::step(MPObjectTable& tbl)
{
	++(*this) ;
	//this->collisionCheck(tbl) ;
	// collision checks are now performed after every object
	// has been moved, not while the objects are being moved
}

void MPObject::push(PVector f)
{
        // th 6.3.00
        acc += f ;
        //
}

void MPObject::rotate(Angle a)
{
	angle += a ;
}

void MPObject::rotate_to(Angle dir)
{
	angle = dir ;
}

void MPObject::dash(Value f)
{
	push(Polar2PVector(f, angle)) ;
}


/*
 *===================================================================
 *Part: PObjectTable
 *===================================================================
 */

PObjectTable::PObjectTable(const TheNumber max)
{
	Max = max ;
	object = new PObject*[Max] ;
	n = 0 ;
}

PObjectTable::~PObjectTable()
{
	delete object ;
}

ostream& operator<< (ostream& o, const PObjectTable& tbl) 
{
	TheNumber i ;

	o << "#PObjectTable[n=" << tbl.n << "/" << tbl.Max << endl ;

	for(i = 0 ; i < tbl.n ; i++)
		o << "  " << *(tbl.object[i]) << endl ;

	return o << "]" << endl ;
}

PObject* PObjectTable::AssignObject(PObject* obj)
{
	object[n] = obj ;
	n++ ;
	return obj ;
}

/* pfr 06/07/200 added short name support */
PObject* PObjectTable::NewObject(const Name nm, const Name s_nm)
{
	PObject *obj = new PObject(nm, s_nm) ;

	AssignObject(obj) ;

	return obj ;
}

/*
 *===================================================================
 *Part: MPObjectTable
 *===================================================================
 */
MPObjectTable::MPObjectTable(const TheNumber max)
{
	randomize() ;
	Max = max ;
	object = new MPObject*[Max] ;
	n = 0 ;
}

MPObjectTable::~MPObjectTable()
{
	delete object ;
}

ostream& operator<< (ostream& o, const MPObjectTable& tbl) 
{
	TheNumber i ;

	o << "#MPObjectTable[n=" << tbl.n << "/" << tbl.Max << endl ;

	 for(i = 0 ; i < tbl.n ; i++)
		o << "\t" << *(tbl.object[i]) << endl ;

	return o << "]" << endl ;
}

MPObject* MPObjectTable::AssignObject(MPObject* obj)
{
	object[n] = obj ;
	n++ ;
	return obj ;
}

/* pfr 06/07/200 added short name support */
MPObject* MPObjectTable::NewObject(const Name nm, const Name s_nm)
{
	MPObject *obj = new MPObject(nm, s_nm) ;

	AssignObject(obj) ;

	return obj ;
}

void MPObjectTable::_inc(  const PlayMode& playmode )
{
	TheNumber i ;
	int sw ;
	MPObject **obj ;
	MPObject *buf ;

	obj = object ;

	/* randomize order */
	for(i = 0; i < n ; i++) {
		sw = irand(n) ;
		buf = *(obj+i) ;
		*(obj+i) = *(obj+sw) ;
		*(obj+sw) = buf ;
	}

	for(i = 0; i < n ; i++) {
		(**obj).step(*this) ;
		obj++ ;
	}

	//new collisions
	collisions ( playmode );
}

void MPObjectTable::collisions ( const PlayMode& playmode )
{
  bool col = false;
  int max = 10;
  int i;

  do 
    {
      col = false;
/*        cout << "max = " << max << endl; */
      for ( i = 0; i < n; ++i )
	{
	  if ( object [ i ] != NULL )
	    {
	       object [ i ]->post_col_pos = PVector ( 0, 0 );
	       object [ i ]->col_count = 0;
	    }
	}

      for ( i = 0; i < n - 1; ++i )
	for ( int j = i + 1; j < n; j++ )
	  {
	  if ( detectCollision ( object [ i ], object [ j ] ) )
	    {
	      col = true;
	      calcCollPos ( object [ i ], object [ j ], playmode );
	    }
	  }

      for ( i = 0; i < n; ++i )
	moveToCollPos ( object [ i ] );

      max--;
    }
  while ( col && max > 0 );

  for ( i = 0; i < n; ++i )
    updateCollVel ( object [ i ] );
}

bool detectCollision ( MPObject* a, MPObject* b )
{
  if ( a != NULL && b != NULL && a != b && a->enable != FALSE && b->enable != FALSE &&
       a->pos.distance ( b->pos ) < a->size + b->size )
    {
      if (a->type == MPO_Ball && b->type == MPO_Player) 
	{
	  ((Player *)b)->alive |= BALL_TO_PLAYER | BALL_COLLIDE;
	  /* we don't want to set last touch if the goalie has caught
	     the ball! */
	  if ( ((Player *)b)->stadium->caught_ball == NEUTRAL )
	    ((Player *)b)->stadium->last_touch = (Player *)b ;
	}
	else if (a->type == MPO_Player && b->type == MPO_Ball) 
	  {
	  ((Player *)a)->alive |= BALL_TO_PLAYER | BALL_COLLIDE;
	  /* we don't want to set last touch if the goalie has caught
	     the ball! */
	  if ( ((Player *)a)->stadium->caught_ball == NEUTRAL )
	    ((Player *)a)->stadium->last_touch = (Player *)a ;
	  }
      else if ( a->type == MPO_Player && b->type == MPO_Player)
	{
	  ((Player *)a)->alive |= PLAYER_COLLIDE ;
	  ((Player *)b)->alive |= PLAYER_COLLIDE ;
	}
      return true;
    }
  return false;
   
}

void calcCollPos ( MPObject* a, MPObject* b, const PlayMode& playmode )
{
  if ( a == NULL || b == NULL )
    return;

  PVector apos;
  PVector bpos;

  if ( playmode != PM_PlayOn  && a->type == MPO_Ball )
    {
      apos = a->pos;
      PVector a2b;
      if ( a->pos == b->pos )
	{
	  Angle b_ang = drand ( -PI, PI );
	  a2b = Polar2PVector ( a->size + b->size, b_ang );
	}
      else
	{
	  a2b = b->pos - a->pos;
	  a2b.r ( a->size + b->size );
	}

      bpos = a->pos + a2b;    
    }
  else if ( playmode != PM_PlayOn && b->type == MPO_Ball )
    {
      bpos = b->pos;
      PVector b2a;
      if ( b->pos == a->pos )
	{
	  Angle a_ang = drand ( -PI, PI );
	  b2a = Polar2PVector ( a->size + b->size, a_ang );
	}
      else
	{
	  b2a = a->pos - b->pos;
	  b2a.r ( a->size + b->size );
	}

      apos = b->pos + b2a;          
    }
  else
    {
      PVector mid = a->pos + b->pos;
      mid /= 2.0;

      PVector mid2a = a->pos - mid;
      PVector mid2b = b->pos - mid;

      if ( a->pos == b->pos )
	{
	  // if the two objects are directly on top on one and other
	  // then they will be sperated at a random angle
	  Angle a_ang = drand ( -PI, PI );
	  Angle b_ang = normalize_angle ( a_ang + PI );
	  mid2a = Polar2PVector ( ( a->size + b->size ) / 2.0 , a_ang );
	  mid2b = Polar2PVector ( ( a->size + b->size ) / 2.0 , b_ang );
	}
      else
	{
	  mid2a.r ( ( a->size + b->size ) / 2.0 );
	  mid2b.r ( ( a->size + b->size ) / 2.0 );
	}

      apos = mid + mid2a;
      bpos = mid + mid2b;

      // 0.01% is added to the movement, as sometimes structural noise
      // means that even though mid2a and mid2b should be 
      // a->size + b->size apart, they are ever so slightly
      // less.
      while ( apos.distance ( bpos ) < a->size + b->size )
	{
	  mid2a.r ( mid2a.r () * 1.0001 );
	  mid2b.r ( mid2b.r () * 1.0001 );
	  apos = mid + mid2a;
	  bpos = mid + mid2b;
	}
    }


/*    cout << "apos = " << a->pos << endl; */
/*    cout << "anewpos = " << (mid + mid2a ) << endl; */

/*    cout << "bpos = " << b->pos << endl; */
/*    cout << "bnewpos = " << (mid + mid2b ) << endl; */

/*    cout << "old distance = " << a->pos.distance ( b->pos ) << endl; */
/*    cout << "new distance = " << apos.distance ( bpos ) << endl; */

  a->post_col_pos += apos;
  b->post_col_pos += bpos;
  a->col_count++;
  b->col_count++;

  a->col = true;
  b->col = true;
   
}

void moveToCollPos ( MPObject* obj )
{
  if ( obj == NULL )
    return;

  if ( obj->col_count > 0 )
    {
/*        cout << "oldpos = " << obj->pos << endl; */
/*        cout << "colpos = " << obj->post_col_pos << endl; */
/*        cout << "colcount = " << obj->col_count << endl; */
      obj->post_col_pos /= obj->col_count;
      obj->pos = obj->post_col_pos;
/*        cout << "newpos = " << obj->pos << endl; */
    }

  obj->post_col_pos = PVector ( 0, 0 );
  obj->col_count = 0;
}

void updateCollVel ( MPObject* obj )
{
  if ( obj == NULL )
    return;

  if ( obj->col )
    {
      obj->vel *= -0.1;
      obj->col = false;
    }
}


void MPObjectTable::_turn()
{
	TheNumber i ;
	int sw ;
	MPObject **obj ;
	MPObject *buf ;

	obj = object ;

	/* randomize order */
	for(i = 0; i < n ; i++) {
		sw = irand(n) ;
		buf = *(obj+i) ;
		*(obj+i) = *(obj+sw) ;
		*(obj+sw) = buf ;
	}

	for(i = 0; i < n ; i++) {
		(**obj)._turn() ;
		obj++ ;
	}
}

/*
 *===================================================================
 *Part: Player & Team
 *===================================================================
 */

Player::Player(Team *tm, Stadium *stad, ID number) : MPObject() 
{
	stadium = stad ;
	team = tm ;
	unum = number ;
	name = new char[128];
	name_far = new char[128];
	name_toofar = new char[128];
	/* th 19.05.00 */
	short_name = new char[128];
	short_name_far = new char[128];
	short_name_toofar = new char[128];
	stamina = stad->server_param->stamina_max ;
	//inertia_moment = stad->server_param->inertia_moment ;
	recovery = 1.0 ;
	//effort = 1.0 ;
	vis_angle = defangle = stad->server_param->visangle ;
	vis_distance = stad->server_param->visdist ;
	vis_send = 4 ;
	sendcnt = 0 ;
	highquality = TRUE ;
	weight = stad->server_param->pweight ;
	weather = &(stad->weather) ;
	//max_speed = stad->server_param->pspeed_max ;
	max_accel = stad->server_param->paccel_max ;
	command_done = FALSE ;
	turn_neck_done = FALSE ;
	offside_mark = FALSE ;
	goalie = FALSE ;
	alive = DISABLE ;
	goalie_catch_ban = FALSE ;
	goalie_catch_probability = stad->server_param->catch_prob ;
	goalie_moves_since_catch = 0;	
	hear_capacity_from_teammate = stad->server_param->hear_max ;
	hear_capacity_from_opponent = stad->server_param->hear_max ;
	pos.x = -(unum * 3 * tm->side) ;
	pos.y = -PITCH_WIDTH/2.0 - 3.0 ;
	angle_neck = 0.0 ;
	angle_neck_committed = 0.0 ;
        angle_body = 0.0 ;

	// pfr 8/14/00: for RC2000 evaluation
	//kick_rand = stad->server_param->kick_rand;
    
	count_kick = count_dash = count_turn = count_say = count_turn_neck = count_move = count_catch = count_change_view = 0 ;
	setPlayerType ( 0 );

	effort = player_type->effortMax ();


}

void Player::init(Value version_tmp, int goalie_flag)
{
	version = version_tmp ;
    goalie = goalie_flag ;
    alive = STAND ;
    version = version_tmp ;
    if (goalie)
        alive |= GOALIE ;
    goalie_catch_ban = FALSE ;
    goalie_moves_since_catch = 0;
    
	sprintf(name,PLAYER_NAME_FORMAT,team->name,unum) ;
	sprintf(name_far,PLAYER_NAME_FAR_FORMAT,team->name) ;
	sprintf(name_toofar,PLAYER_NAME_TOOFAR_FORMAT) ;

	/* th 19.05.00 */
	/* pfr, we'll also show the goalie */
	sprintf(short_name,PLAYER_NAME_FORMAT_SHORT,team->name,unum,
		goalie ? GOALIE_VISUAL_STRING : "") ;
	sprintf(short_name_far,PLAYER_NAME_FAR_FORMAT_SHORT,team->name) ;
	sprintf(short_name_toofar,PLAYER_NAME_TOOFAR_FORMAT_SHORT) ;
	
	// pfr 8/14/00: for RC2000 evaluation
	Value my_prand = stadium->server_param->prand;
	if (stadium->server_param->team_actuator_noise) {
	  my_prand *= team->prand_factor_team;
	  kick_rand *= team->kick_rand_factor_team;
	}

	Set(
		MPO_Player,
		pos,
		//stadium->server_param->psize,
		player_type->playerSize (),
		SideDirection(team->side),
		PVector(0.0,0.0),PVector(0.0,0.0),
		//stadium->server_param->pdecay,
		player_type->playerDecay (),
		my_prand) ;
#ifdef NEW_QSTEP
	dist_qstep_player = team->dist_qstep_team ;
	land_qstep_player = team->land_qstep_team ;
	dir_qstep_player  = team->dir_qstep_team  ;
#endif
}

void Player::rotate(Angle a)
{
	angle_body = normalize_angle( angle + a );
}

void Player::rotate_to(Angle dir)
{
	angle_body = normalize_angle(dir) ;
}

void Player::turn(Value a)
{
	Value r1 = vel.r() ;
	rotate((1.0 + drand(-randp, randp)) * a/(1.0 + inertia_moment * r1)) ;
}

void Player::turn_neck(Value a)
{
	angle_neck = NormalizeNeckAngle( angle_neck + a );
}

void Player::kick(Value power, Angle dir) 
{
	MPObject* ball = stadium->ball ;
	PVector tmp ;
	Value dir_diff ;
	Value dist_ball ;

	alive |= KICK ;

	if (stadium->mode == PM_BeforeKickOff ||
		stadium->mode == PM_AfterGoal_Left ||
		stadium->mode == PM_AfterGoal_Right  ||
		stadium->mode == PM_OffSide_Left ||
		stadium->mode == PM_OffSide_Right )
		return ;

	if(pos.distance(ball->pos) > (player_type->playerSize () + ball->size + player_type->kickableMargin () ) ) {
		alive |= KICK_FAULT ;
		return ;
	}

	if (stadium->server_param->useoffside)
		if (stadium->mark_offside_mark(this))
			return ;

	dir_diff = fabs(this->vangle(*ball)) ;
	tmp = ball->pos - this->pos ;
	dist_ball = tmp.r() - player_type->playerSize () - stadium->server_param->bsize ;

	Value eff_power = power * stadium->server_param->kprate *
			(1 - 0.25*dir_diff/PI - 0.25*dist_ball/player_type->kickableMargin ()) ;

	ball->push(Polar2PVector(eff_power, dir+angle)) ;
	ball->angle = dir+angle ;

	// pfr 8/14/00: for RC2000 evaluation
	// add noise to kick
	Value maxrnd = kick_rand * power / stadium->server_param->maxp;
	PVector kick_noise(drand(-maxrnd,maxrnd),drand(-maxrnd,maxrnd)); 
	//cout << "Kick noise (" << power << "): " << kick_noise << endl;
	ball->push(kick_noise);

	stadium->last_touch = this ;
	stadium->last_kicker = this ;
	stadium->caught_ball = FALSE ;
}

void Player::goalie_catch(Angle dir) 
{
	static RArea p_l(PVector(-PITCH_LENGTH/2+PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	static RArea p_r(PVector(+PITCH_LENGTH/2-PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	static RArea catchable(PVector(stadium->server_param->catch_area_l/2.0, 0.0),
						PVector(stadium->server_param->catch_area_l, stadium->server_param->catch_area_w)) ;
	PVector	tmp ;
	MPObject* ball = stadium->ball ;

	alive |= CATCH ;

	//pfr: we should only be able to catch in PlayOn mode
	if (!this->goalie ||
	    this->goalie_catch_ban ||
	    stadium->mode != PM_PlayOn)
	    /*
	    stadium->mode == PM_BeforeKickOff ||
	    stadium->mode == PM_AfterGoal_Left ||
	    stadium->mode == PM_AfterGoal_Right ||
	    stadium->mode == PM_OffSide_Left ||
	    stadium->mode == PM_OffSide_Right )
	    */
		return ;

	switch (this->team->side) {
	  case LEFT:
		if (!(p_l.inArea(this->pos))) {
			alive |= CATCH_FAULT ;
			return ;
		}
		break ;
 	  case RIGHT:
		if (!(p_r.inArea(this->pos))) {
			alive |= CATCH_FAULT ;
			return ;
		}
		break ;
	}

	this->goalie_catch_ban = stadium->server_param->cban_cycle ;
	tmp = ball->pos - this->pos ;
	tmp.rotate(-(this->angle + dir)) ;
	if (!(catchable.inArea(tmp)) || (drand(1,0) >= goalie_catch_probability)) {
		alive |= CATCH_FAULT ;
		return ;
	}
	relative_ball_ang = ball->pos - this->pos;
	relative_ball_ang /= relative_ball_ang.r();
	goalie_moves_since_catch = 0; //reset the number of times the goalie moved
	stadium->goalie_catch(this) ;
}

Team::Team(Stadium *stad, const Side s)
{
	stadium = stad ;
	enable = FALSE ;
	name = NULLCHAR ;
	side = s ;

	subs = 0;
	for ( int i = 0; i < 7; i++ )
	    ptype_count [ i ] = 0;
	
#ifdef NEW_QSTEP
	if (side == LEFT ){
		dist_qstep_team = stad->server_param->dist_qstep_l ;
		land_qstep_team = stad->server_param->land_qstep_l ;
		dir_qstep_team  = stad->server_param->dir_qstep_l ;
	} else {
		dist_qstep_team = stad->server_param->dist_qstep_r ;
		land_qstep_team = stad->server_param->land_qstep_r ;
		dir_qstep_team  = stad->server_param->dir_qstep_r ;
	}
#endif
	// pfr 8/14/00: for RC2000 evaluation
	if (stad->server_param->team_actuator_noise) {
	  if (side == LEFT ){
	    prand_factor_team     = stad->server_param->prand_factor_l;
	    kick_rand_factor_team = stad->server_param->kick_rand_factor_l;
	  } else {
	    prand_factor_team     = stad->server_param->prand_factor_r;
	    kick_rand_factor_team = stad->server_param->kick_rand_factor_r;
	  }
	}

	Max = MAX_PLAYER ; 
	point = 0 ;
	n = 0 ;
	goalie = FALSE ;
	player = new Player*[Max] ;

//	OnlineCoach* olcoach ;
//	olcoach = new OnlineCoach ;
}

Player *Team::newPlayer(Value version, int goalie_flag) 
{
	if (n >= Max) {
		if (stadium->server_param->verbose)
			cerr << "Warning:Too many players." << endl ;
		return (Player*)NULL ;
	}

	if (goalie && goalie_flag) {
		if (stadium->server_param->verbose)
			cerr << "Warning:Too many goalies." << endl ;
		return (Player*)NULL ;
	}

	if (goalie_flag)
		goalie = TRUE ;

	Player *p = player[n] ;
	n++ ;
	ptype_count [ 0 ] = ptype_count [ 0 ] + 1;
	
	p->init(version, goalie_flag) ;

	return p ;
}


void Player::setPlayerType ( const int& id )
{
  player_type_id = id;
  player_type = stadium->player_types [ id ];
  max_speed = player_type->playerSpeedMax ();
  inertia_moment = player_type->inertiaMoment ();
  decay = player_type->playerDecay ();
  size = player_type->playerSize ();
  kick_rand = player_type->kickRand ();

}


void Player::substitute ( const int& type )
{
    setPlayerType ( type );

    // reset stamina etc.
    stamina =  stadium->server_param->stamina_max ;
    recovery = 1.0 ;
    effort = player_type->effortMax () ;
    hear_capacity_from_teammate =  stadium->server_param->hear_max ;
    hear_capacity_from_opponent =  stadium->server_param->hear_max ;    
}

OnlineCoach::OnlineCoach(Stadium *stad)
  : Coach(stad)
{
  angle = Deg2Rad(90) ;
  side = FALSE ;
  freeform_messages_said = 0;
  freeform_messages_allowed = -1;
  define_messages_left = -1;
  advice_messages_left = -1;
  info_messages_left = -1;
  meta_messages_left = -1;
  msg_left_update_time = -1;
} 

bool OnlineCoach::update_messages_left(TheNumber time)
{
  if (msg_left_update_time >= 0 &&
      time - msg_left_update_time < stadium->server_param->clang_win_size)
    return false;

  define_messages_left = stadium->server_param->clang_define_win;
  meta_messages_left = stadium->server_param->clang_meta_win;
  advice_messages_left = stadium->server_param->clang_advice_win;
  info_messages_left = stadium->server_param->clang_info_win;

  msg_left_update_time = time;

  return true;
}

